@@ -100,7 +100,10 @@ struct _GVfsChannelPrivate
   char reply_buffer[G_VFS_DAEMON_SOCKET_PROTOCOL_REPLY_SIZE];
   int reply_buffer_pos;
   
-  const char *output_data; /* Owned by job */
+  /* output_data is owned by the channel if output_data_free is set,
+   * otherwise it is owned by the job. */
+  const char *output_data;
+  char *output_data_free;
   gsize output_data_size;
   gsize output_data_pos;
 };
@@ -145,8 +148,6 @@ g_vfs_channel_finalize (GObject *object)
   if (channel->priv->backend)
     g_object_unref (channel->priv->backend);
   
-  g_assert (channel->priv->backend_handle == NULL);
-  
   if (G_OBJECT_CLASS (g_vfs_channel_parent_class)->finalize)
     (*G_OBJECT_CLASS (g_vfs_channel_parent_class)->finalize) (object);
 }
@@ -270,6 +271,9 @@ g_vfs_channel_connection_closed (GVfsChannel *channel)
     return;
   channel->priv->connection_closed = TRUE;
   
+  if (g_vfs_backend_get_block_requests (channel->priv->backend))
+    return;
+
   if (channel->priv->current_job == NULL &&
       channel->priv->backend_handle != NULL)
     {
@@ -315,12 +319,22 @@ start_queued_request (GVfsChannel *channel)
 			    channel->priv->queued_requests);
 
       error = NULL;
-      /* This passes on ownership of req->data */
-      job = class->handle_request (channel,
-				   req->command, req->seq_nr,
-				   req->arg1, req->arg2,
-				   req->data, req->data_len,
-				   &error);
+      if (!g_vfs_backend_get_block_requests (channel->priv->backend))
+        {
+	  /* This passes on ownership of req->data */
+	  job = class->handle_request (channel,
+				       req->command, req->seq_nr,
+				       req->arg1, req->arg2,
+				       req->data, req->data_len,
+				       &error);
+        }
+      else
+	{
+	  job = NULL;
+	  error = g_error_new_literal (G_IO_ERROR, G_IO_ERROR_CLOSED,
+	                               _("Channel blocked"));
+	  g_free (req->data);
+	}
 
       if (job != NULL && req->cancelled)
 	{
@@ -360,22 +374,6 @@ got_request (GVfsChannel *channel,
   guint32 command, arg1;
   GList *l;
 
-  if (g_vfs_backend_get_block_requests (channel->priv->backend))
-    {
-      char   *data;
-      gsize   data_len;
-      GError *err = NULL;
-      guint32 seq_nr;
-
-      g_set_error_literal (&err, G_IO_ERROR, G_IO_ERROR_CLOSED,
-			   "Channel blocked");
-      seq_nr = g_ntohl (request->seq_nr);
-      data = g_error_to_daemon_reply (err, seq_nr, &data_len);
-      g_vfs_channel_send_reply (channel, NULL, data, data_len);
-      g_error_free (err);
-      return;
-    }
-
   command = g_ntohl (request->command);
   arg1 = g_ntohl (request->arg1);
 
@@ -604,6 +602,11 @@ send_reply_cb (GObject *source_object,
  error_out:
   
   /* Sent full reply */
+  if (channel->priv->output_data_free)
+    {
+      g_free (channel->priv->output_data_free);
+      channel->priv->output_data_free = NULL;
+    }
   channel->priv->output_data = NULL;
 
   job = channel->priv->current_job;
@@ -675,6 +678,17 @@ g_vfs_channel_send_reply (GVfsChannel *channel,
     }
 }
 
+/* Might be called on an i/o thread */
+void
+g_vfs_channel_send_reply_take (GVfsChannel *channel,
+                               GVfsDaemonSocketProtocolReply *reply,
+                               void *data,
+                               gsize data_len)
+{
+  channel->priv->output_data_free = data;
+  g_vfs_channel_send_reply (channel, reply, data, data_len);
+}
+
 /* Might be called on an i/o thread
  */
 void
@@ -685,7 +699,7 @@ g_vfs_channel_send_error (GVfsChannel *channel,
   gsize data_len;
   
   data = g_error_to_daemon_reply (error, channel->priv->current_job_seq_nr, &data_len);
-  g_vfs_channel_send_reply (channel, NULL, data, data_len);
+  g_vfs_channel_send_reply_take (channel, NULL, data, data_len);
 }
 
 /* Might be called on an i/o thread
@@ -705,7 +719,7 @@ g_vfs_channel_send_info (GVfsChannel *channel,
   reply.arg1 = 0;
   reply.arg2 = g_htonl (data_len);
 
-  g_vfs_channel_send_reply (channel, &reply, data, data_len);
+  g_vfs_channel_send_reply_take (channel, &reply, data, data_len);
 }
 
 int
